/*
  Replacement for the controller in a Vivo Motorized TV Mount, model FD-55
  The Vivo was flakey right out of the box, and then a power failure seems
  to have killed it, as it no longer responds to the remote at all.
  The directions say you can manually get it to move by lifting the
  swing-arm a tiny bit, but not only does this not work, I can see
  no mechanism where it ever worked.

  This uses an Arduino Nano (a 328P in on a small circuit board) to
  replace the dodgy Vivo controller.
  The Vivo runs of 24VDC.  The H-bridge board I have has an on board
  5V regulator, so that should supply the 5V for the Nano.  The radio
  board I ordered might take 12V, so I might need a 7812 for that.

  I'm going to mount a pot in the Vivo case that allows you to set
  the max down position just by turning the pot.  No reason that
  needs to be complicated, nor to be changed every day.

  No, I've decided to save the lower limit in EEPROM.  To set the limit:

  o if you're going to want the TV lower, hold down the 'A' button
      for 2-3 seconds to remove the "user" limit.

  o move tv down to where you want it.

  o hold down the 'B' button for 2-3 seconds to set the new "user"
      limit.

  I've added a current sensor so I can tell if the motor stalls and
  kill the power.  Right now using an ACS712 but it has crappy
  sensitivity for this application, as the motor only draws 60 to
  120mA for a light weight TV.  I'm using a 7lb barbel to simulate
  the weight of my TV, and current draw is about 100mA when lifting
  the TV (60mA when lowering it).

  I have the following shutdown conditions:

  1. top of travel reached -- compile time constant
  2. bottom of travel reached -- compile time constant
  3. user settable lower limit reached -- read from a potentiometer
  4. max run time reached -- compile time constant of 75 seconds
  5. max current exceeded -- compile time constant, about 250mA

  The beep mechanism changed when I realized I couldn't get any decent
  sized speaker in the enclosure.  I've changed to using the beeper,
  cannibalized from the original electronics pcb.  I'll use beep counts
  to indicate different actions.

  For some reason, the button2 library doesn't work with my Arduino Nano.
  Not even the supplied example routine.  So I've written my own light-weight
  button code that works.  Not as flexible, just detects button down and
  reports duration, but good enough for my purposes.

  Bill Dudley
  June 9 2020
*/

#include <EEPROM.h>
#include "EEPROM_rw_anything.h"

// debugging switches:
#define REALLY_MOVE 1
#define VERBOSE_DEBUG 0
#define OVCURRENT_DEBUG 0

// pin definitions:
#define M_POS_SENSOR A6
#define M_CURRENT A7

#define OUTA 2
#define OUTB 3
#define BUTTONA 4
#define BUTTONB 5
#define BEEPER 6
#define ledPin 13

// motor state machine values:
#define STOP 0
#define UP 1
#define DOWN 2

// beeper state machine values:
#define IDLE 0
#define SOUND 1
#define PAUSE 2

#if BuTtOn2
// button press types:
#define NONE 0
#define SHORT 1
#define LONG 2
#define DOUBLE 3
#define VLONG 4
#endif

// constants:
#define MAX_UP 80  	// experimentally determined
#define MAX_DOWN 875	// experimentally determined
#define MAX_TIME 75000L	// approx 75 seconds to move 90 degrees
#define DEBOUNCEMS 50
#define LED_SLOW 1000	// blink rate when STOPped
#define LED_FAST 200
#define FRENETIC 50	// super fast blink to debug overcurrent code
#define OVERCURRENT 250
#define SHORTBEEP 200
#define MEDIUMBEEP 400
#define LONGBEEP 1000
#define STATUS_MSG_SLOW_TIME 4000
#define STATUS_MSG_FAST_TIME 1000	// time between status printf()
#define IIR_EXPONENT 5

// pick one out of all the different current sensors
#define ACS712_05B 0		//  5A AC 185mV/A Vcc 5.0v 2100v isolation *
#define ACS712_20A 0		// 20A AC 100mV/A Vcc 5.0v 2100v isolation
#define ACS712_30A 0		// 30A AC  66mV/A Vcc 5.0v 2100v isolation
#define ACS725_05AB 0		//  5A AC 264mv/A Vcc 3.3v  420v isolation
#define ACS725_10AB 0		// 10A AC 132mv/A Vcc 3.3v  420v isolation
#define ACS725_10AU 0		// 10A DC 264mv/A Vcc 3.3v  420v isolation *
#define ACS725_20AB 0		// 20A AC  66mv/A Vcc 3.3v  420v isolation
#define ACS725_20AU 0		// 20A DC 132mv/A Vcc 3.3v  420v isolation
#define ACS725_30AB 0		// 30A AC  44mv/A Vcc 3.3v  420v isolation
#define ACS725_30AU 0		// 30A DC  88mv/A Vcc 3.3v  420v isolation
#define ACS725_40AB 0		// 40A AC  33mv/A Vcc 3.3v  420v isolation
#define ACS725_50AB 0		// 50A AC  26.4mv/A Vcc 3.3v 420v isolation
#define ACS70331_2P5U3 1	// 2.5A DC 800mv/A Vcc 3.3v 100v isolation
#define ACS70331_2P5B3 0	// 2.5A AC 400mv/A Vcc 3.3v 100v isolation
#define ACS70331_005U3 0	// 5.0A DC 400mv/A Vcc 3.3v 100v isolation
#define ACS70331_005B3 0	// 5.0A AC 200mv/A Vcc 3.3v 100v isolation

// I don't pretend that these constants make for accurate calibration:
#if ACS712_05B
#define COUNTS_TO_mA 17
#elif ACS725_05AB
#define COUNTS_TO_mA 12
#elif ACS725_10AU
#define COUNTS_TO_mA 18
#elif ACS70331_2P5U3
#define COUNTS_TO_mA 7
#elif ACS70331_2P5B3
#define COUNTS_TO_mA 14
#elif ACS70331_005U3
#define COUNTS_TO_mA 14
#elif ACS70331_005B3
#define COUNTS_TO_mA 28
#endif

#if BuTtOn2
Button2 buttonA = Button2(BUTTONA);
Button2 buttonB = Button2(BUTTONB);
byte Astate, Bstate;
#endif

byte Mstate, LEDstate, beepstate, beepcount;
unsigned int Mposition;
unsigned int MCurrentRaw, MCurrentSum, MCurrentAve;
unsigned int IZero;	// current value with motor off
unsigned long MotorRunTime;
int MotorCurrent;
int beepduration, pauseduration;
unsigned int LED_delay;
unsigned long motorstart, lastprint, LEDchange, beepstart;
char buff[100];

#define CURRENT_VERSION 0xdeadbeef

typedef struct config_t {
    unsigned long version;
    unsigned int DownLimit;
} CONFIGBLK;

CONFIGBLK configuration;

typedef struct bTn {
    byte pin;
    byte state;
    byte last_state;
    byte pressed;
    unsigned long down_ms;
    unsigned int down_time_ms;
} BTN;

BTN buttonA, buttonB;

void init_button(BTN *btn, byte pin) {
    btn->pin = pin;
    btn->state = btn->last_state = UP;
    btn->pressed = 0;
    pinMode(btn->pin, INPUT_PULLUP);
}

byte btn_scan(BTN *btn) {
  btn->last_state = btn->state;
  btn->state = digitalRead(btn->pin);
  if(btn->state == 0 && btn->last_state == 1) {	// pressed
    btn->down_ms = millis();
    btn->pressed = 0;
  }
  if(btn->state == 1 && btn->last_state == 0) {	// released
    btn->down_time_ms = millis() - btn->down_ms;
    if(btn->down_time_ms >= DEBOUNCEMS) {
      btn->pressed = 1;
    }
  }
  byte pressed = btn->pressed;
  btn->pressed = 0;
  return pressed;
}

unsigned int btn_time(BTN *btn) {
  return btn->down_time_ms;
}

#if BuTtOn2
// Turns out, the radio unit cannot double click, nor can it short click.
// it can only do long click and very long click.  Minimum click length
// is like 235mS
void handler(Button2& btn) {
    unsigned int time = btn.wasPressedFor();
    switch (btn.getClickType()) {
	case SINGLE_CLICK:
	    if(btn == buttonA) {
	      Astate = SHORT;
#if VERBOSE_DEBUG
	      Serial.print(F("AS: "));
#endif
	    }
	    else {
	      Bstate = SHORT;
#if VERBOSE_DEBUG
	      Serial.print(F("BS: "));
#endif
	    }
#if VERBOSE_DEBUG
	    Serial.println(time);
#endif
	    break;
	case DOUBLE_CLICK:
	    if(btn == buttonA) {
	      Astate = DOUBLE;
#if VERBOSE_DEBUG
	      Serial.println(F("AD"));
#endif
	    }
	    else {
	      Bstate = DOUBLE;
#if VERBOSE_DEBUG
	      Serial.println(F("BD"));
#endif
	    }
	    break;
	case LONG_CLICK:
	    if(btn == buttonA) {
	      Astate = (time > 2000) ? VLONG : LONG ;
#if VERBOSE_DEBUG
	      Serial.print(F("AL: "));
#endif
	    }
	    else {
	      Bstate = (time > 2000) ? VLONG : LONG ;
#if VERBOSE_DEBUG
	      Serial.print(F("BL: "));
#endif
	    }
#if VERBOSE_DEBUG
	    Serial.println(time);
#endif
	    break;
    }
}
#endif

void TVmove(byte direction) {
  if(direction == STOP) {
    digitalWrite(OUTA, 1);
    digitalWrite(OUTB, 1);
    MotorRunTime = MAX_TIME;
  }
  else if(direction == UP) {
    digitalWrite(OUTA, 1);
#if REALLY_MOVE
    digitalWrite(OUTB, 0);
#endif
  }
  else if(direction == DOWN) {
#if REALLY_MOVE
    digitalWrite(OUTA, 0);
#endif
    digitalWrite(OUTB, 1);
  }
  Mstate = direction;
  motorstart = millis();
}

void setup() {
  Serial.begin(38400);
  pinMode(ledPin, OUTPUT);
  pinMode(OUTA, OUTPUT);
  pinMode(OUTB, OUTPUT);
  pinMode(BEEPER, OUTPUT);
  digitalWrite(BEEPER, LOW);
  init_button(&buttonA, BUTTONA);
  init_button(&buttonB, BUTTONB);
  // digitalWrite(BUTTONA, HIGH);
  // digitalWrite(BUTTONB, HIGH);  // enable pullup
  TVmove(STOP);
#if BuTtOn2
  buttonA.setClickHandler(handler);
  buttonA.setLongClickHandler(handler);
  buttonA.setDoubleClickHandler(handler);
  buttonB.setClickHandler(handler);
  buttonB.setLongClickHandler(handler);
  buttonB.setDoubleClickHandler(handler);
#endif
  EEPROM_readAnything(0, configuration);
  if(configuration.version != CURRENT_VERSION) {
    Serial.println(F("version wrong, initializing configuration"));
    configuration.version = CURRENT_VERSION;
    configuration.DownLimit = MAX_DOWN;
    EEPROM_writeAnything(0, configuration);
  }
  sprintf(buff, "version = %ld, config.dl = %d\n",
    configuration.version, configuration.DownLimit);
  Serial.print(buff);
  if(configuration.DownLimit < 300) {
    configuration.DownLimit = MAX_DOWN;
    EEPROM_writeAnything(0, configuration);
    sprintf(buff, "fixed bad config.dl = %d\n", configuration.DownLimit);
    Serial.print(buff);
  }
  beepcount = 1;	// beep on startup
  beepduration = SHORTBEEP;
  LEDstate = 0;
#if BuTtOn2
  Astate = NONE;
  Bstate = NONE;
#endif
  lastprint = 0L;
  LEDchange = 0L;
  LED_delay = LED_SLOW;
  IZero = 440;
}

void loop() {
static int fc;
int status_msg_time;
  Mposition = analogRead(M_POS_SENSOR);
  MCurrentRaw = analogRead(M_CURRENT);
#if no_OVCURRENT_DEBUG
  if(++fc >= 1600) {
    sprintf(buff, "MCurrentRaw = %d\n", MCurrentRaw);
    Serial.print(buff);
    fc = 0;
  }
#endif
  if(MCurrentAve == 0) {
    MCurrentAve = MCurrentRaw;
#if IIR_EXPONENT
    MCurrentSum = MCurrentAve << IIR_EXPONENT;
#endif
  }
  else {	// IIR filter
#if IIR_EXPONENT
    MCurrentSum -= MCurrentSum >> IIR_EXPONENT;
    MCurrentSum += MCurrentRaw;
    MCurrentAve = MCurrentSum >> IIR_EXPONENT;
#else
    MCurrentAve = MCurrentRaw;
#endif
    if(Mstate == STOP) {
      IZero = MCurrentAve;
    }
  }
  MotorCurrent = (MCurrentAve - IZero) * COUNTS_TO_mA;
#if VERBOSE_DEBUG
  status_msg_time = (Mstate == STOP) ? STATUS_MSG_SLOW_TIME : STATUS_MSG_FAST_TIME ;
  if(lastprint + status_msg_time < millis()) {
    sprintf(buff, "config.DownLimit = %d, Mstate = %s, mpos = %d, Iraw = %d,  Iave = %d, Iz = %d, CTmA = %d, I = %d\n",
      configuration.DownLimit,
      (Mstate == STOP) ? "stop" : (Mstate == UP) ? "up" : "down",
      Mposition,
      MCurrentRaw, MCurrentAve, IZero, COUNTS_TO_mA, MotorCurrent);
    Serial.print(buff);
    lastprint = millis();
  }
#endif
#if BuTtOn2
  // buttonA.loop();
  // if(Astate)
#endif
  if(btn_scan(&buttonA))
  {
#if BuTtOn2
#if VERBOSE_DEBUG
    sprintf(buff, "(1) As = %s\n",
      (Astate==SHORT) ? "short" : (Astate == VLONG) ? "vlong" : (Astate==LONG) ? "long" : (Astate == DOUBLE) ? "double" : "none");
    Serial.print(buff);
#endif
    // if(Astate == VLONG)
#endif
#if VERBOSE_DEBUG
    sprintf(buff, "A down time = %dms", btn_time(&buttonA));
    Serial.println(buff);
#endif
    if(btn_time(&buttonA) > 2000)
    {
#if VERBOSE_DEBUG
      Serial.println(F("long click, save user limit if reasonable."));
#endif
      if(Mposition > 300 && Mposition < MAX_DOWN) {
	configuration.DownLimit = Mposition;
	Serial.println(F("saving DownLimit"));
	EEPROM_writeAnything(0, configuration);
	beepcount = 3;
	beepduration = LONGBEEP;
      }
    }
    else {
      if(Mstate == STOP) {
	if((Mposition < (MAX_DOWN - 2))
	&& (Mposition < (configuration.DownLimit - 2))) {
	  // IZero = MCurrentAve;
	  TVmove(DOWN);
	  beepcount = 1;
	  beepduration = SHORTBEEP;
#if VERBOSE_DEBUG
	  Serial.println(F("move down"));
#endif
	}
      }
      else {	// if moving, stop moving
	TVmove(STOP);
	beepcount = 2;
	beepduration = SHORTBEEP;
#if VERBOSE_DEBUG
	Serial.println(F("stop moving (A)"));
#endif
      }
    }
#if BuTtOn2
    Astate = NONE;
#endif
  }
#if BuTtOn2
  // buttonB.loop();
  // if(Bstate)
#endif
  if(btn_scan(&buttonB))
  {
#if BuTtOn2
#if VERBOSE_DEBUG
    sprintf(buff, "(2) Bs = %s\n",
      (Bstate==SHORT) ? "short" : (Bstate == VLONG) ? "vlong" : (Bstate==LONG) ? "long" : (Bstate == DOUBLE) ? "double" : "none");
    Serial.print(buff);
#endif
    // if(Bstate == VLONG)
#endif
#if VERBOSE_DEBUG
    sprintf(buff, "B down time = %dms", btn_time(&buttonB));
    Serial.println(buff);
#endif
    if(btn_time(&buttonB) > 2000)
    {
      configuration.DownLimit = MAX_DOWN;
      EEPROM_writeAnything(0, configuration);
      Serial.println(F("reset DownLimit to max"));
      beepcount = 2;
      beepduration = LONGBEEP;
    }
    else {
      if(Mstate == STOP) {
	if(Mposition > MAX_UP+5) {
	  // IZero = MCurrentAve;
	  TVmove(UP);
	  beepcount = 1;
	  beepduration = SHORTBEEP;
#if VERBOSE_DEBUG
	  Serial.println(F("move up"));
#endif
	}
      }
      else {	// if moving, stop moving
	TVmove(STOP);
	beepcount = 2;
	beepduration = SHORTBEEP;
#if VERBOSE_DEBUG
	Serial.println(F("stop moving (B)"));
#endif
      }
    }
#if BuTtOn2
    Bstate = NONE;
#endif
  }
  if(Mstate == STOP || LED_delay != FRENETIC) {
    LED_delay = Mstate ? LED_FAST * Mstate : LED_SLOW ;
  }
#if OVCURRENT_DEBUG
  if((motorstart + 1000 < millis())
  && MotorCurrent > OVERCURRENT) {
    sprintf(buff, "motor current > 200 = %d\n", MotorCurrent);
    Serial.print(buff);
    LED_delay = FRENETIC;
  }
#endif
  if((Mstate != STOP && motorstart + MotorRunTime < millis())
  || (Mstate == UP && Mposition < MAX_UP)
  || (Mstate == DOWN && Mposition > MAX_DOWN)
  || (Mstate == DOWN && Mposition > configuration.DownLimit)
  || ((motorstart + 1000 < millis()) && MotorCurrent > OVERCURRENT)
  ) {
#define ERROR_DEBUG 1
#if ERROR_DEBUG
    Serial.print(F("stop moving: "));
    if(motorstart + MotorRunTime < millis()) {
      Serial.println(F("run time"));
    }
    if(Mstate == UP && Mposition < MAX_UP) {
      Serial.println(F("UP limit"));
    }
    if(Mstate == DOWN && Mposition > MAX_DOWN) {
      Serial.println(F("MAX DOWN limit"));
    }
    if(Mstate == DOWN && Mposition > configuration.DownLimit) {
      Serial.println(F("USER DOWN limit"));
    }
    if((motorstart + 1000 < millis()) && MotorCurrent > OVERCURRENT) {
      Serial.println(F("over current"));
    }
#endif
    TVmove(STOP);
    beepcount = 1;
    beepduration = SHORTBEEP;
  }
  if(LEDchange + LED_delay < millis()) {
    digitalWrite(ledPin, LEDstate);
    LEDstate ^= 1;
    LEDchange = millis();
  }
  if(beepcount) {
    switch(beepstate) {
      case IDLE :
Play:	digitalWrite(BEEPER, HIGH);
	beepstart = millis();
	beepstate = SOUND;
	beepcount--;
	break;
      case SOUND :
	if(beepstart + (unsigned long)beepduration < millis()) {
	  digitalWrite(BEEPER, LOW);
	  if(beepcount > 0) {
	    beepstate = PAUSE;
	    pauseduration = SHORTBEEP;
	    beepstart = millis();
	  }
	  else {
	    beepstate = IDLE ;
	  }
	}
	break;
      case PAUSE :
	if(beepstart + (unsigned long)pauseduration < millis()) {
	  if(beepcount > 0) {
	    goto Play;
	  }
	  else {	// this should never happen
	    beepstate = IDLE ;
	  }
	}
	break;
    }
  }
  else {
    if(beepstart + (unsigned long)beepduration < millis()) {
      digitalWrite(BEEPER, LOW);
      beepstate = IDLE ;
    }
  }
}
